@@ -31,7 +31,7 @@ from fresco.util.urls import url_join, make_query
 from fresco.util.http import parse_header
 from fresco.util.wsgi import unicode_to_environ
 
-__version__ = '7.4.1'
+__version__ = '7.4.2'
 
 #: Registry for xpath multimethods
 xpath_registry = {}
@@ -78,7 +78,7 @@ class XPathMultiMethod(object):
         for xpath, func in self.endpoints:
             if el in xpath(el):
                 return func(*args, **kwargs)
-        raise NotImplementedError("Function %s not implemented for element %r"\
+        raise NotImplementedError("Function %s not implemented for element %r"
                                   % (self.__name__, el,))
 
     def register(self, xpath, func):
@@ -94,12 +94,9 @@ class XPathMultiMethod(object):
         # Note that ".. comment block" is required to fool rst/sphinx into
         # correctly parsing the indented paragraphs when there is only one
         # registered endpoint.
-        doc = 'For elements matching ``%s``:'\
-              '\n%s\n\n.. comment block\n\n' % (
+        doc = 'For elements matching ``%s``:n%s\n\n.. comment block\n\n' % (
                     xpath,
-                    '\n'.join('    %s' % line
-                              for line in func_doc.split('\n'))
-        )
+                  '\n'.join('    %s' % line for line in func_doc.split('\n')))
         self.__doc__ += doc
         self.__name__ = func.__name__
 
@@ -390,7 +387,7 @@ class ElementWrapper(object):
         group)
         """
         for el in self.el.xpath(
-            "./ancestor-or-self::form[1]"\
+            "./ancestor-or-self::form[1]"
             "//input[@type='radio' and @name=$name]",
             name=self.el.attrib.get('name', '')
         ):
@@ -432,9 +429,8 @@ class ElementWrapper(object):
         previously selected option.
         """
         if 'multiple' not in \
-           self.el.xpath('./ancestor-or-self::select[1]')[0].attrib:
-            for el in \
-            self.el.xpath("./ancestor-or-self::select[1]//option"):
+                self.el.xpath('./ancestor-or-self::select[1]')[0].attrib:
+            for el in self.el.xpath("./ancestor-or-self::select[1]//option"):
                 if 'selected' in el.attrib:
                     del el.attrib['selected']
 
@@ -593,8 +589,8 @@ class ElementWrapper(object):
                 elif 'checked' in el.attrib:
                     del el.attrib['checked']
             if found != values:
-                raise AssertionError("Values %r not present"\
-                                     " in checkbox group %r" %\
+                raise AssertionError("Values %r not present"
+                                     " in checkbox group %r" %
                                      (values - found,
                                       self.el.attrib.get('name')))
 
@@ -619,8 +615,8 @@ class ElementWrapper(object):
             elif 'checked' in el.attrib:
                 del el.attrib['checked']
         if value is not None and not found:
-            raise AssertionError("Value %r not present"\
-                                 " in radio button group %r" %\
+            raise AssertionError("Value %r not present"
+                                 " in radio button group %r" %
                                  (value, self.el.attrib.get('name')))
         return self
 
@@ -709,7 +705,7 @@ class ElementWrapper(object):
             elif 'selected' in opt.attrib:
                 del opt.attrib['selected']
         if not found and value is not None:
-            raise AssertionError("Value %r not present in select %r" %\
+            raise AssertionError("Value %r not present in select %r" %
                                  (value, self.el.attrib.get('name')))
         return self
 
@@ -853,7 +849,8 @@ class TestAgent(object):
           requests for the WSGI application and return a new ``TestAgent``
           object
 
-        - ``request``, ``response`` - the `Fresco <http://fresco.redgecko.org/>`_
+        - ``request``, ``response`` - the
+          `Fresco <http://fresco.redgecko.org/>`_
           request and response objects associated with the last WSGI request.
 
         - ``body`` - the body response as a bytes object
@@ -896,7 +893,9 @@ class TestAgent(object):
 
     def __init__(self, app, request=None, response=None, cookies=None,
                  history=None, validate_wsgi=True, host='127.0.0.1',
-                 port='8080', loglevel=None, logger=None, close_response=True):
+                 port='8080', loglevel=None, logger=None,
+                 environ_defaults=None,
+                 close_response=True):
         """
         Initialize the ``TestAgent`` object.
 
@@ -918,6 +917,7 @@ class TestAgent(object):
 
         The following parameters are intended for internal use only:
 
+        :param environ_defaults: A list of default WSGI environ values
         :param history: The history list
         :param cookie: The cookie store to use for the TestAgent.
         :param request: The ``fresco.request.Request`` object associated with
@@ -975,7 +975,8 @@ class TestAgent(object):
         # Stores file upload field values in forms
         self.file_uploads = {}
 
-        self.environ_defaults = self.__class__.environ_defaults.copy()
+        self.environ_defaults = (environ_defaults or self.environ_defaults)\
+                                    .copy()
         self.environ_defaults.update({'SERVER_NAME': host,
                                       'SERVER_PORT': str(port)})
 
@@ -997,8 +998,8 @@ class TestAgent(object):
 
         if '?' in PATH_INFO:
             if 'QUERY_STRING' in kwargs:
-                raise AssertionError("QUERY_STRING specified both in "\
-                                     "PATH_INFO and as argument to "\
+                raise AssertionError("QUERY_STRING specified both in "
+                                     "PATH_INFO and as argument to "
                                      "make_environ")
             PATH_INFO, querystring = PATH_INFO.split('?', 1)
             kwargs['QUERY_STRING'] = unicode_to_environ(querystring)
@@ -1103,6 +1104,7 @@ class TestAgent(object):
             response,
             self.cookies,
             history,
+            environ_defaults=self.environ_defaults,
             **self.options
         )
         if self.options['logger']:
@@ -1121,34 +1123,28 @@ class TestAgent(object):
             return agent.follow_all()
         return agent
 
-    def get(self, PATH_INFO='/', data=None, charset='UTF-8', follow=True,
-            history=True, check_status=True, **kwargs):
+    def _get(self, PATH_INFO='/', data=None, charset='UTF-8', follow=True,
+             history=True, check_status=True, method='GET', **kwargs):
         """
         Make a GET request to the application and return the response.
 
-        PATH_INFO
+        :param PATH_INFO:
             The path to request from the application. This must be URL encoded.
             As a convenience, query string data can be appended to this.
             Example value: ``agent.get('/my%20script?param=1')``
 
-        data
-            Query string data to be appended to the request, must be either a
-            dict or list of ``(name, value)`` tuples.
+        :param data: Query string data to be appended to the request, must be
+                     either a dict or list of ``(name, value)`` tuples.
 
-        charset
-            Encoding used for any unicode values encountered in ``data``
+        :charset: Encoding used for any unicode values encountered in ``data``
 
-        follow
-            If false, redirect responses will not be followed
+        :follow: If false, redirect responses will not be followed
 
-        history
-            If true, a new entry will be added to the history attribute for the
-            resulting TestAgent object
-
-        check_status
-            If true, any non success HTTP status code will result in an
-            AssertionError being raised
+        :history: If true, a new entry will be added to the history attribute
+                  for the resulting TestAgent object
 
+        :check_status: If true, any non success HTTP status code will result in
+                       an AssertionError being raised
         """
         if data is not None:
             kwargs.setdefault('QUERY_STRING', make_query(data,
@@ -1169,33 +1165,35 @@ class TestAgent(object):
             check_status,
         )
 
-    def post(self, PATH_INFO='/', data=None, charset='UTF-8', follow=True,
-             history=True, check_status=True, **kwargs):
+    def _post(self, PATH_INFO='/', data=None, charset='UTF-8', follow=True,
+              history=True, check_status=True,
+              content_type='application/x-www-form-urlencoded', method='POST',
+              **kwargs):
         """
         Make a POST request to the application and return the response.
 
-        PATH_INFO
-            The path to request from the application. This must be URL encoded.
+        :param PATH_INFO: The path to request from the application. This must
+                          be URL encoded.
 
-        data
-            POST data to be sent to the application, must be either a
-            dict or list of ``(name, value)`` tuples.
+        :param data: POST data to be sent to the application. Can be a byte
+                     string of the raw post data, a dict or a list of ``(name,
+                     value)`` tuples.
 
-        charset
-            Encoding used for any unicode values encountered in ``data``
+        :param charset: Encoding used for any unicode values encountered in
+                        ``data``
 
-        follow
-            If false, redirect responses will not be followed
+        :param content_type: The content type header for the posted data. The
+                             default is good for testing form submissions, if
+                             you want to test an API you may need to change
+                             this to something else, eg 'application/json'
 
-        history
-            If true, a new entry will be added to the history attribute for the
-            resulting TestAgent object
-
-        check_status
-            If true, any non success HTTP status code will result in an
-            AssertionError being raised
+        :param follow: If false, redirect responses will not be followed
 
+        :param history: If true, a new entry will be added to the history
+                        attribute for the resulting TestAgent object
 
+        :param check_status: If true, any non success HTTP status code will
+                             result in an AssertionError being raised
         """
         if data is None:
             data = []
@@ -1204,15 +1202,16 @@ class TestAgent(object):
             PATH_INFO = url_join_same_server(self.request.url,
                                              PATH_INFO)
 
-        data = make_query(data, charset=charset).encode('ASCII')
+        if not isinstance(data, bytes):
+            data = make_query(data, charset=charset).encode('ASCII')
         wsgi_input = BytesIO(data)
         wsgi_input.seek(0)
 
         return self._request(
             self.make_environ(
-                'POST',
+                method,
                 PATH_INFO,
-                CONTENT_TYPE="application/x-www-form-urlencoded",
+                CONTENT_TYPE=content_type,
                 CONTENT_LENGTH=str(len(data)),
                 wsgi_input=wsgi_input,
                 **kwargs
@@ -1222,6 +1221,21 @@ class TestAgent(object):
             check_status,
         )
 
+    get = _get
+    post = _post
+
+    def head(self, *args, **kwargs):
+        return self._get(method='HEAD', *args, **kwargs)
+
+    def delete(self, *args, **kwargs):
+        return self._get(method='DELETE', *args, **kwargs)
+
+    def put(self, *args, **kwargs):
+        return self._post(method='PUT', *args, **kwargs)
+
+    def patch(self, *args, **kwargs):
+        return self._post(method='PATCH', *args, **kwargs)
+
     def post_multipart(self, PATH_INFO='/', data=None, files=None,
                        charset='UTF-8', follow=True, history=True,
                        check_status=True, **kwargs):
@@ -1284,10 +1298,8 @@ class TestAgent(object):
             else:
                 if not isinstance(value, bytes):
                     value = value.encode(charset)
-                headers = [
-                    ('Content-Disposition',
-                    'form-data; name="%s"' % (key,))
-                ]
+                headers = [('Content-Disposition',
+                            'form-data; name="%s"' % (key,))]
                 return headers, value
 
         items = chain(
@@ -1344,8 +1356,8 @@ class TestAgent(object):
         refilled on the reloaded page, provided that the reloaded page has
         exactly the same fields present in the same order.
         """
-        self.options['logger'] and self.options['logger'].debug("Reload: %s",
-                                          self.request.path)
+        if self.options['logger']:
+            self.options['logger'].debug("Reload: %s", self.request.path)
         env = self.request.environ.copy()
         del env['fresco.request']
 
@@ -1741,7 +1753,7 @@ def url_join_same_server(baseurl, url):
     url = url_join(baseurl, url)
     url = urlparse(url)
     if urlparse(baseurl)[:2] != url[:2]:
-        raise ValueError("URI links to another server: %s" %\
+        raise ValueError("URI links to another server: %s" %
                          (urlunparse(url),))
     return urlunparse(('', '') + url[2:])
 
